'''
Created on Nov 2, 2012

@author: vcarlos
'''
import math
import random
from mem.memory import Memory



#===============================================================================
# Cache
#===============================================================================
class Cache(Memory):
    '''
    Implements a cache behaviour
    '''


    def __init__(self, size, main_memory_size, energy_per_read, energy_per_write):
        '''
        Constructor
        '''
        super(Cache, self).__init__(size, energy_per_read)
        self.size = size
        self.main_memory_size = main_memory_size
        self.number_of_address_bits = int(math.log(main_memory_size, 2))
        self.energy_per_read = energy_per_read
        self.energy_per_write = energy_per_write
        
        self.data = {}
        
        
#===============================================================================
# DMCache
#===============================================================================
class DMCache(Cache):
    
    def __init__(self, number_words, main_memory_size, size, mapping, energy_per_read, energy_per_write):
        super(DMCache, self).__init__(size, main_memory_size, energy_per_read, energy_per_write)
        self.mapping = mapping
        self.number_words = number_words
        
        self._str_format = '{0:0' + str(self.number_of_address_bits) + 'b}'
        
        

    def _MapAddress(self, address):
        # format the address  to binary like string
        address = self._str_format.format(address)
        
        # where we have ones, are those that form the address in the cache
        mapped_address = ""
        for i in xrange(self.number_of_address_bits):
            if self.mapping[i] == "1":
                mapped_address += address[i]
            
        # return the integer (address on cache)
#        print "mapped_address", mapped_address
        return int(mapped_address, 2)
    
    
    def _ReadFromCache(self, full_address, cache_address):
        
        value = full_address
#        print full_address, cache_address
        if self.data.has_key(cache_address):
            value = self.data[cache_address]
            v = value[-1]
            if v != full_address:
#                print cache_address, v, full_address
#                print bin(cache_address), bin(v), bin(full_address)
                value.append(full_address)
                self.data[cache_address] = value
                return 0
            else:
                return 1
        else:
            # address not there yet, cold miss
            self.data[cache_address] = [value]
            return 0
            
        return 0
    
    def _ReadFromCache2(self, full_address, cache_address):
        
        value = full_address
        if self.data.has_key(cache_address):
            if self.data[cache_address] == full_address:
                return 1
            else:
                self.data[cache_address] = value
                return 0
        else:
            # address not there yet, cold miss
            self.data[cache_address] = value
            return 0
            
        
    def Read(self, address):
        '''
            "Read" from memory and return 1 if hit, 0 if miss.
        '''
        pos = self._MapAddress(address)
#        print hex(pos)
#        if pos != 0: 
#            print"0x%0.2x" % pos
        hit = self._ReadFromCache(address, pos)
#        if hit:
#            print"0x%0.2x" % pos
#            print"0x%0.5x" % address
#            print
            
        return hit
    
    
    def GetTotalMissConflicts(self):
        total = 0
        for item in self.data.iteritems():
            # in each position there is a list with all addresses mapped to that position
            conflicting_addresses = item[1]
            conflicting_addresses.sort()
            
            j = 0
            number_of_addresses = len(conflicting_addresses)
            if number_of_addresses > 1:
        #        print conflicting_addresses
                while j < number_of_addresses:
                    address = conflicting_addresses[j]
                    c = conflicting_addresses.count(address)
                    total += c - 1
                    j += c
    
        return total
    
    
#===============================================================================
# PhasedDMCache
#===============================================================================
class PhasedDMCache(DMCache):
    
    def __init__(self, number_words, main_memory_size, size, mapping, energy_per_read_tag, energy_per_read_data, energy_per_write_tag, energy_per_write_data):
        super(PhasedDMCache, self).__init__(number_words, main_memory_size, size, mapping, energy_per_read_tag+energy_per_read_data, energy_per_write_tag+energy_per_write_data)
        
        self.energy_per_read_tag = energy_per_read_tag
        self.energy_per_read_data = energy_per_read_data 
        self.energy_per_write_tag = energy_per_write_tag
        self.energy_per_write_data = energy_per_write_data 
        
    
    def GetEnergyPerTagRead(self):
        return self.energy_per_read_tag
    
    def GetEnergyPerDataRead(self):
        return self.energy_per_read_data
    
    def GetEnergyPerWrite(self):
        return self.energy_per_write_tag + self.energy_per_write_data        
        
    
    
#===============================================================================
# XORDMCache
#===============================================================================
class XORDMCache(DMCache):
    
    
    def __init__(self, main_memory_size, size, mapping, energy_per_read, energy_per_write):
        super(XORDMCache, self).__init__(main_memory_size, size, mapping, energy_per_read, energy_per_write)
        
        
    
    def _MapAddress(self, address):
        
        
        # format the address  to binary like string
        address = self._str_format.format(address)
        
        bits = ""
        for i in range(3, 9):
            bits += str(int(address[-1 * i + 1]) ^ int(address[-1 * i - 6]))
           
#------------------------------------------------------------ reverse
        a = bits
        a = list(a)
        a.reverse()
        bits = ""
        for i in a:
            bits += i
            
#        print bits
        pos = int(bits + address[-2:], 2)
        
#        print bits + address[-2:]
#        print pos
        return pos
    
    
#===============================================================================
# PhasedNWayCache
#===============================================================================
class PhasedNWayCache(DMCache):
    
    def __init__(self, number_words, main_memory_size, size, mapping, number_of_ways, energy_per_read_tag, energy_per_read_data, energy_per_write_tag, energy_per_write_data):
        super(PhasedNWayCache, self).__init__(number_words, main_memory_size, size, mapping, energy_per_read_tag+energy_per_read_data, energy_per_write_tag+energy_per_write_data)
        
        self.number_of_ways = number_of_ways
        self.size = size
        
        self.number_of_accesses = 0
        
        self.energy_per_read_tag = energy_per_read_tag
        self.energy_per_read_data = energy_per_read_data 
        self.energy_per_write_tag = energy_per_write_tag
        self.energy_per_write_data = energy_per_write_data 
        
        self.mapping = mapping
        self._str_format = '{0:0' + str(self.number_of_address_bits) + 'b}'
        
        # list of dicts is the data structure. each dict is a way
        self.data = []
        for i in xrange(number_of_ways):
            self.data.append(dict())
            
    
    def _ReadFromWay(self, way, full_address, cache_address):
        # if this line is already present in the cache, check for its contents
        # if not, it is a miss
        if way.has_key(cache_address):
            value = way[cache_address]
            # if the contents match, it is a hit, otherwise a miss
            return value == full_address
        else:
            return 0
    
    def _ReplaceCacheLine(self, full_address, cache_address):
        # so there was a miss, need to put the data somewhere, 
        # therefore a random replace policy is used to do so
#        which_way = random.randint(0, self.number_of_ways-1)
        which_way = self.number_of_accesses % self.number_of_ways
        self.data[which_way][cache_address] = full_address
    
    
    
    def Read(self, address):
        self.number_of_accesses += 1
        
        # first get the cache line where the data might be found
        pos = self._MapAddress(address)
        
        # now read from each way and check for a hit. 
        # if there is a hit, just return it
        for each_way in self.data:
            hit = self._ReadFromWay(each_way, address, pos)
            if hit:
                return 1
            
        # if it got to this point, a miss happened
        # so it is necessary to determine in which way the data will be store (cache replacement)
        self._ReplaceCacheLine(address, pos)
        
        # return miss
        return 0
        
    
    def GetEnergyPerTagRead(self):
        return self.energy_per_read_tag * self.number_of_ways
    
    def GetEnergyPerDataRead(self):
        return self.energy_per_read_data
    
    def GetEnergyPerWrite(self):
        return self.energy_per_write_tag + self.energy_per_write_data
    
    
    
#===============================================================================
# WayHaltingCache
#===============================================================================
#TODO: CHECK constructor before testing
class WayHaltingCache(DMCache):
    def __init__(self, number_words, main_memory_size, size, mapping, number_of_ways, 
                 energy_per_read_halt, energy_per_read, 
                 energy_per_write_halt, energy_per_write, 
                 number_of_tag_bits, number_of_halt_tag_bits):
            super(WayHaltingCache, self).__init__(number_words, main_memory_size, size, mapping, energy_per_read, energy_per_write)
            
            self.number_of_ways = number_of_ways
            
            self.number_of_index_bits = int(math.log(size/number_of_ways/4, 2))
            print "self.number_of_index_bits", self.number_of_index_bits
            
            self.energy_per_read_halt = energy_per_read_halt
            self.energy_per_write_halt = energy_per_write_halt
            
            self.mapping = mapping
            self._str_format = '{0:0' + str(self.number_of_address_bits) + 'b}'
            
            self.number_of_accesses = 0
            self.number_of_hits = 0
            self.number_of_haltings = 0
            self.total_number_of_haltings = 0
            
            self.number_of_tag_bits = number_of_tag_bits
            self.number_of_halt_tag_bits = number_of_halt_tag_bits
            
            self.total_ways = 0
            
            # list of pair of dicts is the data structure. each dict is a way
            # the first dict is the halt array and the second is the rest of the address array
            self.data = []
            for i in xrange(number_of_ways):
                self.data.append((dict(), dict()))
            
            
    
    def _ReadFromWay(self, way, full_address, cache_address):
        # if this line is already present in the cache, check for its contents
        # if not, it is a miss
        if way[1].has_key(cache_address):
            value = way[1][cache_address]
            # if the contents match, it is a hit, otherwise a miss
            return value == full_address
        else:
            return 0
    
    def _ReplaceCacheLine(self, full_address, cache_address, halt_tag):
        # so there was a miss, need to put the data somewhere, 
        # therefore a random replace policy is used to do so
#        which_way = random.randint(0, self.number_of_ways-1)
        which_way = self.number_of_accesses % self.number_of_ways
        self.data[which_way][0][cache_address] = halt_tag
        self.data[which_way][1][cache_address] = full_address
    
    
    def _CheckWayHalting(self, halt_tag, cache_address):
#        print "halt_tag", halt_tag
        list_of_ways = []
        way_n = 0
        for each_way in self.data:
            halt_array = each_way[0]
            for v in halt_array.itervalues():
                if v == halt_tag:
#                    print 'halt_tag', halt_tag, v
#                    print "v", v
#                    print "way_n", way_n
                    list_of_ways.append(each_way)
                    break
            way_n += 1
#            if halt_array.has_key(cache_address):
#                if halt_array[cache_address] == halt_tag:
#                    list_of_ways.append(each_way)
                
        return list_of_ways
    
                    
    def Read(self, address):
        
        self.number_of_accesses += 1
        
        # first get the cache line where the data might be found
        pos = self._MapAddress(address)
#        print 'pos', pos
        
        # split the address into tag, halt tag and index
        bin_address = self._str_format.format(address)
        
        tag = bin_address[:self.number_of_tag_bits]
        halt_tag = tag[-self.number_of_halt_tag_bits:]
#        print bin_address
#        print tag
#        print halt_tag

        # now check the halt arrays to see which ones should be read from (the ones not present are disabled=halted)
        list_of_ways = self._CheckWayHalting(int(halt_tag, 2), pos)

        self.total_ways += len(list_of_ways)
        
        self.number_of_haltings = self.number_of_ways - len(list_of_ways)
        
#        print "self.number_of_haltings",  self.number_of_haltings
        
        self.total_number_of_haltings += self.number_of_haltings 
         
        # if there is a hit, just return it
        for each_way in list_of_ways:
            hit = self._ReadFromWay(each_way, address, pos)
            if hit:
                return 1
            
        # if it got to this point, a miss happened
        # so it is necessary to determine in which way the data will be store (cache replacement)
        self._ReplaceCacheLine(address, pos, int(halt_tag, 2))
        
        # return miss
        return 0
        
    
    def GetEnergyPerRead(self):
        return self.energy_per_read_halt * self.number_of_ways + self.energy_per_read * (self.number_of_ways - self.number_of_haltings)
    
    def GetEnergyPerWrite(self):
        return self.energy_per_write_halt + self.energy_per_write    
    
    
    def __str__(self, *args, **kwargs):
        print_ways = []
        for way in self.data:
            pway = []
            for i in xrange(16):
                pway.append((0,0))
            keys = way[0].keys()
            keys.sort()
            for k in keys:
                tag = way[1][k]
                halt = way[0][k]
                cmb = tag , halt
                pway[k] = (cmb)
            print_ways.append(pway)
        
#        print print_ways
        
        for i in xrange(16):
            try:
                t,h = print_ways[0][i]
                str_format = '{0:0' + str(self.number_of_halt_tag_bits) + 'b}'
                way0 = self._str_format.format(t) + " " + str_format.format(h)
                
#                .format(print_ways[0][i])
#                way0 = '{0:0' + str(self.number_of_halt_tag_bits) + 'b}'.format((print_ways[0][i]))
            except IndexError:
                way0 = "0".center(self.number_of_address_bits + self.number_of_halt_tag_bits + 1)
            try:
                t,h = print_ways[1][i]
                str_format = '{0:0' + str(self.number_of_halt_tag_bits) + 'b}'
                way1 = self._str_format.format(t) + " " + str_format.format(h)                
                
                
#                way1 = '{0:0' + str(self.number_of_halt_tag_bits) + 'b}'.format((print_ways[1][i]))
            except:
                way1 = "1".center(self.number_of_address_bits + self.number_of_halt_tag_bits + 1)
            try:
                t,h = print_ways[2][i]
                str_format = '{0:0' + str(self.number_of_halt_tag_bits) + 'b}'
                way2 = self._str_format.format(t) + " " + str_format.format(h)                
                
#                way2 = str(print_ways[2][i]).center(self.number_of_address_bits + self.number_of_halt_tag_bits + 1)
            except:
                way2 = "2".center(self.number_of_address_bits + self.number_of_halt_tag_bits + 1)
            try:
                t,h = print_ways[3][i]
                str_format = '{0:0' + str(self.number_of_halt_tag_bits) + 'b}'
                way3 = self._str_format.format(t) + " " + str_format.format(h)                
                
                
#                way3 = str(print_ways[3][i]).center(self.number_of_address_bits + self.number_of_halt_tag_bits + 1)
            except:
                way3 = "3".center(self.number_of_address_bits + self.number_of_halt_tag_bits + 1)
        
            print way0 + "   " + way1 + "   " + way2 + "   " + way3 
            
        return ""        
        
            
                
            
    
    
#===============================================================================
# Main
#===============================================================================
if __name__ == '__main__':
    pass    
