//`define MAX_REGISTER_FILE_SIZE 10
//`define MAX_SRAM_SIZE 13
//`define USE_REGISTER_FILE

module DataMemory #(
  parameter NB_ADDRESS = 5,
  parameter NB_DATA    = 32
  )
  (
  input clk,
  
  // Read port
  input logic [NB_ADDRESS-1:0] addrRead,
  input logic ceRead,
  input logic re,
  output logic [NB_DATA-1:0] dataOutRead,
  
  // Write port
  input logic [NB_ADDRESS-1:0] addrWrite,
  input logic [NB_DATA-1:0] dataInWrite,
  input logic we,
  input logic ceWrite
  );

`ifdef USE_REGISTER_FILE  
  localparam MAX_NB_ADDRESS_BITS = 10;
`else
  localparam MAX_NB_ADDRESS_BITS = 13;
`endif

  localparam NB_BANKS = fu_NBSubbanks(NB_ADDRESS, MAX_NB_ADDRESS_BITS);
  localparam NB_EXTRA_ADDRESS_BITS = fu_ExtraAddressBits(NB_ADDRESS, MAX_NB_ADDRESS_BITS);
  localparam MSB_ADDRESS_BIT = fu_MSBAddressBit(NB_ADDRESS, MAX_NB_ADDRESS_BITS);
  
  // then I have as many ce as banks (one bit for each bank)
  logic [NB_BANKS-1:0] ceR, ceW;
  // and as many outputs from the banks as the number of banks
  logic [NB_DATA-1:0] dataOut [NB_BANKS-1:0];
  
  integer j;
  
  
`ifdef USE_VLOG_MEM
  BasicMemorySync # (
      .NB_ADDRESS (NB_ADDRESS),
      .NB_DATA (NB_DATA)
    )
    u_BasicMemory (
      .clk (clk),
      // read port
      .addrRead (addrRead),
      .ceRead (ceRead),
      .re (re),
      .dataOutRead (dataOutRead),
      
      // write port
      .addrWrite (addrWrite),
      .dataInWrite (dataInWrite),
      .ceWrite (ceWrite),
      .we (we)
     );
`else

  generate
    case (NB_ADDRESS)
`ifdef USE_REGISTER_FILE
      5: DP_RF_32x32 u (
           .QA (dataOut[0]),
           .AA (addrRead),
           .CLKA (clk),
           .CENA (~ceRead | re),
           .AB (addrWrite),
           .DB (dataInWrite),
           .CLKB (clk),
           .CENB (~(ceWrite & we))
        );
      6: DP_RF_64x32 u (
           .QA (dataOut[0]),
           .AA (addrRead),
           .CLKA (clk),
           .CENA (~ceRead | re),
           .AB (addrWrite),
           .DB (dataInWrite),
           .CLKB (clk),
           .CENB (~(ceWrite & we))
        );
      7: DP_RF_128x32 u (
           .QA (dataOut[0]),
           .AA (addrRead),
           .CLKA (clk),
           .CENA (~ceRead | re),
           .AB (addrWrite),
           .DB (dataInWrite),
           .CLKB (clk),
           .CENB (~(ceWrite & we))
        );
      8: DP_RF_256x32 u (
           .QA (dataOut[0]),
           .AA (addrRead),
           .CLKA (clk),
           .CENA (~ceRead | re),
           .AB (addrWrite),
           .DB (dataInWrite),
           .CLKB (clk),
           .CENB (~(ceWrite & we))
        );
      9: DP_RF_512x32 u (
           .QA (dataOut[0]),
           .AA (addrRead),
           .CLKA (clk),
           .CENA (~ceRead | re),
           .AB (addrWrite),
           .DB (dataInWrite),
           .CLKB (clk),
           .CENB (~(ceWrite & we))
        );
      10: DP_RF_1024x32 u (
           .QA (dataOut[0]),
           .AA (addrRead),
           .CLKA (clk),
           .CENA (~ceRead | re),
           .AB (addrWrite),
           .DB (dataInWrite),
           .CLKB (clk),
           .CENB (~(ceWrite & we))
        );
      11: begin
        
        // instantiations
        
        DP_RF_1024x32 u1 (
           .QA (dataOut[0]),
           .AA (addrRead),
           .CLKA (clk),
           .CENA (ceR[0] & ~re),
           .AB (addrWrite),
           .DB (dataInWrite),
           .CLKB (clk),
           .CENB (ceW[0] & we)
        );
                
        DP_RF_1024x32 u2 (
           .QA (dataOut[1]),
           .AA (addrRead),
           .CLKA (clk),
           .CENA (ceR[1] & ~re),
           .AB (addrWrite),
           .DB (dataInWrite),
           .CLKB (clk),
           .CENB (ceW[1] & we)
        );        
        
        end
      12: begin
        
        // instantiations
        
        DP_RF_1024x32 u1 (
           .QA (dataOut[0]),
           .AA (addrRead),
           .CLKA (clk),
           .CENA (ceR[0] & ~re),
           .AB (addrWrite),
           .DB (dataInWrite),
           .CLKB (clk),
           .CENB (ceW[0] & we)
        );
                
        DP_RF_1024x32 u2 (
           .QA (dataOut[1]),
           .AA (addrRead),
           .CLKA (clk),
           .CENA (ceR[1] & ~re),
           .AB (addrWrite),
           .DB (dataInWrite),
           .CLKB (clk),
           .CENB (ceW[1] & we)
        );  
                
        DP_RF_1024x32 u3 (
           .QA (dataOut[2]),
           .AA (addrRead),
           .CLKA (clk),
           .CENA (ceR[2] & ~re),
           .AB (addrWrite),
           .DB (dataInWrite),
           .CLKB (clk),
           .CENB (ceW[2] & we)
        );  
                
        DP_RF_1024x32 u4 (
           .QA (dataOut[3]),
           .AA (addrRead),
           .CLKA (clk),
           .CENA (ceR[3] & ~re),
           .AB (addrWrite),
           .DB (dataInWrite),
           .CLKB (clk),
           .CENB (ceW[3] & we)
        );  
        
        end      
      13: begin
        
        // instantiations
        DP_RF_1024x32 u1 (
           .QA (dataOut[0]),
           .AA (addrRead),
           .CLKA (clk),
           .CENA (ceR[0] & ~re),
           .AB (addrWrite),
           .DB (dataInWrite),
           .CLKB (clk),
           .CENB (ceW[0] & we)
        );
                
        DP_RF_1024x32 u2 (
           .QA (dataOut[1]),
           .AA (addrRead),
           .CLKA (clk),
           .CENA (ceR[1] & ~re),
           .AB (addrWrite),
           .DB (dataInWrite),
           .CLKB (clk),
           .CENB (ceW[1] & we)
        );  
                
        DP_RF_1024x32 u3 (
           .QA (dataOut[2]),
           .AA (addrRead),
           .CLKA (clk),
           .CENA (ceR[2] & ~re),
           .AB (addrWrite),
           .DB (dataInWrite),
           .CLKB (clk),
           .CENB (ceW[2] & we)
        );  
                
        DP_RF_1024x32 u4 (
           .QA (dataOut[3]),
           .AA (addrRead),
           .CLKA (clk),
           .CENA (ceR[3] & ~re),
           .AB (addrWrite),
           .DB (dataInWrite),
           .CLKB (clk),
           .CENB (ceW[3] & we)
        );  
        
        
        DP_RF_1024x32 u5 (
           .QA (dataOut[4]),
           .AA (addrRead),
           .CLKA (clk),
           .CENA (ceR[4] & ~re),
           .AB (addrWrite),
           .DB (dataInWrite),
           .CLKB (clk),
           .CENB (ceW[4] & we)
        );
                
        DP_RF_1024x32 u6 (
           .QA (dataOut[5]),
           .AA (addrRead),
           .CLKA (clk),
           .CENA (ceR[5] & ~re),
           .AB (addrWrite),
           .DB (dataInWrite),
           .CLKB (clk),
           .CENB (ceW[5] & we)
        );  
                
        DP_RF_1024x32 u7 (
           .QA (dataOut[6]),
           .AA (addrRead),
           .CLKA (clk),
           .CENA (ceR[6] & ~re),
           .AB (addrWrite),
           .DB (dataInWrite),
           .CLKB (clk),
           .CENB (ceW[6] & we)
        );  
                
        DP_RF_1024x32 u8 (
           .QA (dataOut[7]),
           .AA (addrRead),
           .CLKA (clk),
           .CENA (ceR[7] & ~re),
           .AB (addrWrite),
           .DB (dataInWrite),
           .CLKB (clk),
           .CENB (ceW[7] & we)
        );                  
        end    
        
`else
      5: DP_SRAM_32x32 u (
           .QA(dataOut[0]),
           .CLKA(clk),
           .CENA(~ceRead),
           .WENA(~re),
           .AA(addrRead),
           .DA({32{1'b0}}),
           .OENA(1'b0),
           .QB(),
           .CLKB(clk),
           .CENB(~ceWrite),
           .WENB(~we),
           .AB(addrWrite),
           .DB(dataInWrite),
           .OENB(1'b1)
      );
      6: DP_SRAM_64x32 u (
           .QA(dataOut[0]),
           .CLKA(clk),
           .CENA(~ceRead),
           .WENA(~re),
           .AA(addrRead),
           .DA({32{1'b0}}),
           .OENA(1'b0),
           .QB(),
           .CLKB(clk),
           .CENB(~ceWrite),
           .WENB(~we),
           .AB(addrWrite),
           .DB(dataInWrite),
           .OENB(1'b0)
      );
      7: DP_SRAM_128x32 u (
           .QA(dataOut[0]),
           .CLKA(clk),
           .CENA(~ceRead),
           .WENA(~re),
           .AA(addrRead),
           .DA({32{1'b0}}),
           .OENA(1'b0),
           .QB(),
           .CLKB(clk),
           .CENB(~ceWrite),
           .WENB(~we),
           .AB(addrWrite),
           .DB(dataInWrite),
           .OENB(1'b0)
      );
      8: DP_SRAM_256x32 u (
           .QA(dataOut[0]),
           .CLKA(clk),
           .CENA(~ceRead),
           .WENA(~re),
           .AA(addrRead),
           .DA({32{1'b0}}),
           .OENA(1'b0),
           .QB(),
           .CLKB(clk),
           .CENB(~ceWrite),
           .WENB(~we),
           .AB(addrWrite),
           .DB(dataInWrite),
           .OENB(1'b0)
      );
      9: DP_SRAM_512x32 u (
           .QA(dataOut[0]),
           .CLKA(clk),
           .CENA(~ceRead),
           .WENA(~re),
           .AA(addrRead),
           .DA({32{1'b0}}),
           .OENA(1'b0),
           .QB(),
           .CLKB(clk),
           .CENB(~ceWrite),
           .WENB(~we),
           .AB(addrWrite),
           .DB(dataInWrite),
           .OENB(1'b0)
      );
      10: DP_SRAM_1024x32 u (
           .QA(dataOut[0]),
           .CLKA(clk),
           .CENA(~ceRead),
           .WENA(~re),
           .AA(addrRead),
           .DA({32{1'b0}}),
           .OENA(1'b0),
           .QB(),
           .CLKB(clk),
           .CENB(~ceWrite),
           .WENB(~we),
           .AB(addrWrite),
           .DB(dataInWrite),
           .OENB(1'b0)
      );
      11: DP_SRAM_2048x32 u (
           .QA(dataOut[0]),
           .CLKA(clk),
           .CENA(~ceRead),
           .WENA(~re),
           .AA(addrRead),
           .DA({32{1'b0}}),
           .OENA(1'b0),
           .QB(),
           .CLKB(clk),
           .CENB(~ceWrite),
           .WENB(~we),
           .AB(addrWrite),
           .DB(dataInWrite),
           .OENB(1'b0)
      );
      12: DP_SRAM_4096x32 u (
           .QA(dataOut[0]),
           .CLKA(clk),
           .CENA(~ceRead),
           .WENA(~re),
           .AA(addrRead),
           .DA({32{1'b0}}),
           .OENA(1'b0),
           .QB(),
           .CLKB(clk),
           .CENB(~ceWrite),
           .WENB(~we),
           .AB(addrWrite),
           .DB(dataInWrite),
           .OENB(1'b0)
      );
      13: DP_SRAM_8192x32 u (
           .QA(dataOut[0]),
           .CLKA(clk),
           .CENA(~ceRead),
           .WENA(~re),
           .AA(addrRead),
           .DA({32{1'b0}}),
           .OENA(1'b0),
           .QB(),
           .CLKB(clk),
           .CENB(~ceWrite),
           .WENB(~we),
           .AB(addrWrite),
           .DB(dataInWrite),
           .OENB(1'b0)
      );
      14: begin
        
        // instantiations
        DP_SRAM_8192x32 u1 (
             .QA(dataOut[0]),
             .CLKA(clk),
             .CENA(~ceR[0]),
             .WENA(~re),
             .AA(addrRead),
             .DA({32{1'b0}}),
             .OENA(1'b0),
             .QB(),
             .CLKB(clk),
             .CENB(~ceW[0]),
             .WENB(~we),
             .AB(addrWrite),
             .DB(dataInWrite),
             .OENB(1'b0)
        );        
        DP_SRAM_8192x32 u2 (
             .QA(dataOut[1]),
             .CLKA(clk),
             .CENA(~ceR[0]),
             .WENA(~re),
             .AA(addrRead),
             .DA({32{1'b0}}),
             .OENA(1'b0),
             .QB(),
             .CLKB(clk),
             .CENB(~ceW[0]),
             .WENB(~we),
             .AB(addrWrite),
             .DB(dataInWrite),
             .OENB(1'b0)
        );        
        end
 
`endif      
    endcase
      
  endgenerate
  
  
  // input logic
  always_comb 
    begin
      if (NB_BANKS != 1)// begin
//        ceR[0] = ceRead;
//        ceW[0] = ceWrite;
//      end
//      else
        for (j = 0; j < NB_BANKS; j++) begin
          if (addrRead[NB_ADDRESS-1:NB_ADDRESS-NB_EXTRA_ADDRESS_BITS-1] == j) 
            ceR[j] = ceRead;
          else
            ceR[j] = 1'b0;
            
          if (addrWrite[NB_ADDRESS-1:NB_ADDRESS-NB_EXTRA_ADDRESS_BITS-1] == j) 
            ceW[j] = ceWrite;
          else
            ceW[j] = 1'b0;
        end
    end
    
  // output logic
  always_comb 
    begin
      if (NB_BANKS == 1)
        dataOutRead = dataOut[0];
      else
        dataOutRead = dataOut[addrRead[NB_ADDRESS-1:NB_ADDRESS-NB_EXTRA_ADDRESS_BITS-1]];
    end   
    
`endif    
  
  // compilation time
  function automatic int fu_ExtraAddressBits (int nb_address, int max_nb_address_bits);
    int temp;
    temp = (nb_address - max_nb_address_bits);
    fu_ExtraAddressBits = (temp <= 0) ? 0 : temp;
  endfunction
  
  
  function automatic int fu_NBSubbanks (int nb_address, int max_nb_address_bits);
    fu_NBSubbanks = 2 ** fu_ExtraAddressBits(nb_address, max_nb_address_bits);
  endfunction   
  
  // compilation time
  function automatic int fu_MSBAddressBit (int nb_address, int max_nb_address_bits);
    int extra_bits = fu_ExtraAddressBits(nb_address, max_nb_address_bits);
    if (extra_bits == 0)
      fu_MSBAddressBit = nb_address;
    else    
      fu_MSBAddressBit = max_nb_address_bits;
  endfunction   
    
endmodule 
